"""Functions to read the challenge score weight file for the GCP pipeline."""

import numpy as np
import helper_code


def load_table(table_file):  # pylint:disable=g-doc-args
  """Load a table with row and column names.

  The table should have the following form:

  ,    a,   b,   c
  a, 1.2, 2.3, 3.4
  b, 4.5, 5.6, 6.7
  c, 7.8, 8.9, 9.0

  Args:
    table_file: Path to the table file.

  Returns:
    A tuple of (rows, cols, values) for the table.

  Raises:
    Exception: if table is empty or has rows with different lengths.
  """
  table = list()
  with open(table_file, 'r') as f:
    for i, l in enumerate(f):
      arrs = [arr.strip() for arr in l.split(',')]
      table.append(arrs)

  # Define the numbers of rows and columns and check for errors.
  num_rows = len(table) - 1
  if num_rows < 1:
    raise Exception('The table {} is empty.'.format(table_file))  # pylint:disable=g-doc-exception
  row_lengths = set(len(table[i]) - 1 for i in range(num_rows))
  if len(row_lengths) != 1:
    raise Exception(  # pylint:disable=g-doc-exception
        'The table {} has rows with different lengths.'.format(table_file))
  num_cols = min(row_lengths)
  if num_cols < 1:
    raise Exception('The table {} is empty.'.format(table_file))  # pylint:disable=g-doc-exception

  # Find the row and column labels.
  rows = [table[0][j + 1] for j in range(num_rows)]
  cols = [table[i + 1][0] for i in range(num_cols)]

  # Find the entries of the table.
  values = np.zeros((num_rows, num_cols), dtype=np.float64)
  for i in range(num_rows):
    for j in range(num_cols):
      value = table[i + 1][j + 1]
      if helper_code.is_finite_number(value):
        values[i, j] = float(value)
      else:
        values[i, j] = float('nan')

  return rows, cols, values


def load_weights(weight_file):
  """Load the table with the weight matrix."""
  rows, cols, values = load_table(weight_file)

  # Split the equivalent classes.
  rows = [set(row.split('|')) for row in rows]
  cols = [set(col.split('|')) for col in cols]
  assert rows == cols

  # Identify the classes and the weight matrix.
  classes = rows
  weights = values

  return classes, weights


