"""This is a libarary that must be included in CinC submission, and we are allowed to modify.

The functions in this library are called by train_model.py and test_model.py .
"""

import json
import os
import time
import numpy as np
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf  # pylint:disable=g-import-not-at-top
from joblib import Parallel, delayed

import helper_code
import load_weights_gcp
import model_autoencoder
import preprocess

twelve_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF', 'V1', 'V2', 'V3', 'V4',
                'V5', 'V6')
six_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF')
four_leads = ('I', 'II', 'III', 'V2')
three_leads = ('I', 'II', 'V2')
two_leads = ('I', 'II')
lead_sets = (twelve_leads, six_leads, four_leads, three_leads, two_leads)

NUM_JOBS_FOR_IO = 8
FILTER_AND_SCALE = True
WINDOW_SIZE = 1024
WINDOW_SAMPLE_MIN = 300

# Model hyper-parameters
EMBEDDING_DIM = 64

# Optimal number of iterations.
ENCODER_LR = 0.0001
CLASSIFIER_LR = 0.0001
ENCODER_EPOCHS_OPT = 10
CLASSIFIER_EPOCHS_OPT = 15
CLASSIFIER_EXTRA_EPOCHS_OPT = 15

BATCH_SIZE = 16
RETRAIN_ENCODER = True
USE_AGE = False
USE_SEX = False
RECORD_PREDICT_AGGREGATION = 'mean'  # {'mean', 'max'}
IMBALANCE_STRATEGY = 'loss'  # {'bias', 'loss'}

THRESHOLD_OPTIMIZATION = 'none'  # {'cinc', 'independent', 'none'}
THRESHOLD_OPT = 0.035

TRAIN_SINGLE_MODEL = False

################################################################################
#
# Training model function
#
################################################################################


# Train your model. This function is *required*. You should edit this function
# to add your code, but do *not* change the arguments of this function.
def _read_train_data(header_file, recording_file):
  header = helper_code.load_header(header_file)
  recording = helper_code.load_recording(recording_file)
  ecg = preprocess_data(header, recording)
  if ecg is not None:
    if FILTER_AND_SCALE:
      ecg = preprocess.filter_and_scale(ecg)
    ecg_list = preprocess.window_data(ecg, WINDOW_SIZE, WINDOW_SAMPLE_MIN, FILTER_AND_SCALE, False)
    return ecg_list

def training_code(data_directory, model_directory):
  """Find header and recording files."""
  t0 = time.perf_counter()

  header_files, recording_files = helper_code.find_challenge_files(
      data_directory)
  num_recordings = len(recording_files)
  print(f'find {num_recordings} files; {int(time.perf_counter() - t0)} sec')

  if not num_recordings:
    raise Exception('No data was provided.')

  # Create a folder for the model if it does not already exist.
  if not os.path.isdir(model_directory):
    os.mkdir(model_directory)

  # Extract the data.
  if NUM_JOBS_FOR_IO == 1:
    record_list = []
    for i in range(num_recordings):
      if i % 1000 == 0:
        print(f'{i} files extracted: {int(time.perf_counter() - t0)} sec')
      header = helper_code.load_header(header_files[i])
      recording = helper_code.load_recording(recording_files[i])
      ecg = preprocess_data(header, recording)
      if ecg is None:
        continue
      if FILTER_AND_SCALE:
        ecg = preprocess.filter_and_scale(ecg)
      ecg_list = preprocess.window_data(
              ecg, WINDOW_SIZE, WINDOW_SAMPLE_MIN, FILTER_AND_SCALE, False)
      record_list.extend(ecg_list)
  else:
    ecg_lists = Parallel(n_jobs=NUM_JOBS_FOR_IO)(
            delayed(_read_train_data)(hf, rf) for (hf, rf) in zip(header_files, recording_files))
    record_list = []
    for x in ecg_lists:
      record_list.extend(x)
  print(f'Data extracted ({len(record_list)}): {int(time.perf_counter() - t0)} sec')
  
  # Read classes and weights used for evaluation.
  eval_classes, eval_weights = load_weights_gcp.load_weights('weights.csv')
  eval_classes_flat = preprocess.flatten_classes(eval_classes)
  print(f'Read weights: {int(time.perf_counter() - t0)} sec')

  # Create training data dictionary.
  data = {}
  data['age'] = np.zeros(len(record_list), dtype='float32')
  data['sex'] = [[] for i in range(len(record_list))]
  data['idnum'] = [[] for i in range(len(record_list))]
  data['label'] = np.zeros((len(record_list), len(eval_classes_flat)))
  for i, rec in enumerate(record_list):
    for label in rec['label']:
      if label in eval_classes_flat:
        data['label'][i, eval_classes_flat.index(label)] = 1
    data['age'][i] = rec['age']
    data['sex'][i] = rec['sex']
    data['idnum'][i] = rec['idnum']
  data['sex'] = np.array(data['sex'])
  data['idnum'] = np.array(data['idnum'])
  print(f'Data dictionary created: {int(time.perf_counter() - t0)} sec')
  # Loop over each lead set.
  for leads in lead_sets:
    print(f'Training model for {len(leads)}-lead set: {", ".join(leads)}...')
    if (not TRAIN_SINGLE_MODEL) or len(leads) == 12:
      data['ecg'] = np.zeros((len(record_list), WINDOW_SIZE, len(leads)),
                             dtype='float32')
      for i, rec in enumerate(record_list):
        rec = preprocess.combine_leads_dict(rec, leads)
        data['ecg'][i] = rec['ecg']
      # Shuffle training data.
      data_shuffled = shuffle_data(data, len(record_list))
      # Train the classifier on all data.
      classifier, threshold_opt = model_autoencoder.fit_model_gcp(
          data=data_shuffled,
          classes=eval_classes,
          weights=eval_weights,
          input_shape=(WINDOW_SIZE, len(leads)),
          embedding_dim=EMBEDDING_DIM,
          autoencoder_epochs=ENCODER_EPOCHS_OPT,
          classifier_epochs=CLASSIFIER_EPOCHS_OPT,
          classifier_extra_epochs=CLASSIFIER_EXTRA_EPOCHS_OPT,
          encoder_lr=ENCODER_LR,
          classifier_lr=CLASSIFIER_LR,
          batch_size=BATCH_SIZE,
          threshold_optimization=THRESHOLD_OPTIMIZATION,
          use_age=USE_AGE,
          use_sex=USE_SEX,
          retrain_encoder=RETRAIN_ENCODER,
          imbalance_strategy=IMBALANCE_STRATEGY,
          start_time=t0)
      if THRESHOLD_OPTIMIZATION == 'none':
        threshold_opt = THRESHOLD_OPT
    # Save the model.
    save_model(model_directory, leads, eval_classes_flat, classifier,
               threshold_opt)
    print(f'Model and params saved: {int(time.perf_counter() - t0)} sec')


################################################################################
#
# Running trained model function
#
################################################################################


# Run your trained model. This function is *required*. You should edit this
# function to add your code, but do *not* change the arguments of this function.
def run_model(model, header, recording):
  """Runs the trained model and saves predictions."""
  classes = model['classes']
  leads = model['leads']
  classifier = model['classifier']
  assert len(leads) == recording.shape[0]

  # Preprocess
  long_record = preprocess_data(header, recording)

  # Filter and scale (optional)
  if FILTER_AND_SCALE:
    long_record = preprocess.filter_and_scale(long_record)

  # Window long records.
  record_list = preprocess.window_data(
      long_record, WINDOW_SIZE, WINDOW_SAMPLE_MIN, FILTER_AND_SCALE, False)

  # Predict
  y_pred_list = {c: [] for c in classes}
  for record in record_list:
    record = preprocess.combine_leads_dict(record, leads)
    if TRAIN_SINGLE_MODEL:
      rec_ecg = np.zeros((record['ecg'].shape[0], 12))
      rec_ecg[:, np.isin(twelve_leads, leads)] = record['ecg']
      record['ecg'] = rec_ecg
    # Remove model-irrelevant keys and format.
    record = {key: record[key] for key in ['ecg', 'age', 'sex']}
    record['ecg'] = np.expand_dims(record['ecg'], 0).astype('float32')
    record['age'] = np.expand_dims(np.array([record['age']]), 0)
    record['sex'] = np.expand_dims(np.array([record['sex']]), 0)
    y_pred = classifier.predict(record)
    for c in classes:
      y_pred_list[c].append(y_pred[c])

  # Calculate probabilities and predictions.
  if RECORD_PREDICT_AGGREGATION == 'max':
    probabilities_dict = {c: np.max(v) for c, v in y_pred_list.items()}
  elif RECORD_PREDICT_AGGREGATION == 'mean':
    probabilities_dict = {c: np.mean(v) for c, v in y_pred_list.items()}
  else:
    raise ValueError('Unknown record aggregation method.')
  probabilities = np.array([probabilities_dict[c] for c in classes])
  if isinstance(model['threshold_opt'], dict):
    labels = [(probabilities_dict[c] > model['threshold_opt'][c]).astype(int)
              for c in classes]
  else:
    labels = [(probabilities_dict[c] > model['threshold_opt']).astype(int)
              for c in classes]

  return classes, labels, probabilities


################################################################################
#
# File I/O functions
#
################################################################################


def save_model(model_directory, leads, classes, classifier, threshold_opt):
  """Saves a trained model."""
  d = {
      'leads': leads,
      'classes': classes,
      'threshold_opt': threshold_opt,
  }
  filename = os.path.join(model_directory, get_model_filename(leads))
  with open(filename + '.params', 'w') as fp:
    json.dump(d, fp)
  classifier.save(filename, save_format='tf')


# Load a trained model. This function is *required*. You should edit this
# function to add your code, but do *not* change the arguments of this function.
def load_model(model_directory, leads):
  """Loads a tensorflow model."""
  filename = os.path.join(model_directory, get_model_filename(leads))
  with open(filename + '.params', 'r') as fp:
    params = json.load(fp)
  classifier = tf.keras.models.load_model(filename)
  return {'classifier': classifier, 'threshold_opt': params['threshold_opt'],
          'classes': params['classes'], 'leads': params['leads']}


# Define the filename(s) for the trained models. This function is not required.
# You can change or remove it.
def get_model_filename(leads):
  sorted_leads = helper_code.sort_leads(leads)
  return 'model_' + '-'.join(sorted_leads)


def preprocess_data(header, recording):
  """Preprocessing header and recordings and put everything in a dictionary."""
  try:
    leadnames = helper_code.get_leads(header)
    data = preprocess.preprocess_ecg({
        'samprate': helper_code.get_frequency(header),
        'baselines': helper_code.get_baselines(header, leadnames),
        'num_samples': helper_code.get_num_samples(header),
        'leadnames': leadnames,
        'age': helper_code.get_age(header),
        'sex': helper_code.get_sex(header),
        'label': helper_code.get_labels(header),
        'idnum': helper_code.get_recording_id(header),
    }, recording)
  except:  # pylint:disable=bare-except
    print('Could not extract data')
    data = None
  return data


def shuffle_data(data, n_record):
  """Shuffles training data."""
  idx_shuffle = np.arange(n_record)
  np.random.shuffle(idx_shuffle)
  data_shuffled = {}
  for key in data:
    data_shuffled[key] = data[key][idx_shuffle]
  return data_shuffled
